/* $LAAS: viamMainCodels.c 2011/04/11 11:59:31 mallet $ */

/*
 * Copyright (c) 2007-2008,2010-2011 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice,  this list of  conditions and the following disclaimer in
 *      the  documentation  and/or  other   materials provided  with  the
 *      distribution.
 *
 * THIS  SOFTWARE IS PROVIDED BY  THE  COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND  ANY  EXPRESS OR IMPLIED  WARRANTIES,  INCLUDING,  BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES  OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR  PURPOSE ARE DISCLAIMED. IN  NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR      CONTRIBUTORS  BE LIABLE FOR   ANY    DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF  SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE   OF THIS SOFTWARE, EVEN   IF ADVISED OF   THE POSSIBILITY OF SUCH
 * DAMAGE.
 *
 *                                      Anthony Mallet on Tue Jan 23 2007
 */

#include <portLib.h>
#include <stdio.h>
#include <sys/time.h>

#include <viam/viamlib.h>

#include "server/viamHeader.h"
#include "shared.h"


/* --- local data ---------------------------------------------------- */

viam_handle_t vh;		/**< viam handle */
static viam_bank_t *configured;	/**< array of configured banks */
static int nconfigured;


/* --- viamMainInit -------------------------------------------------- */

/** \brief Initialization codel (fIDS, ...)
 *
 * \return OK or ERROR
 */

STATUS
viamMainInit(int *report)
{
  SDI_F->init_done = 0;
  nconfigured = 0;
  configured = NULL;
  vh = viam_init();

#ifdef VIAM_DISPLAY
  if (viamDisplayInit(report) != OK) return ERROR;
#endif
  return viamAcqInit(report);
}


/* --- viamMainEnd --------------------------------------------------- */

/** \brief Termination codel of Main execution task
 *
 * \return OK or ERROR
 */

STATUS
viamMainEnd(void)
{
  viam_bankdata_t *d;
  int i;

  SDI_F->init_done = 0;

  viamAcqEnd();

  for(i=0; i<nconfigured; i++) {
    d = viam_bank_usrdata(configured[i]);
    if (d->pdata) viamPosterDelete(vh, d->pdata);
  }
  if (configured) free(configured);
  nconfigured = 0;
  configured = NULL;

  viam_hardware_detach(vh);
  viam_release(vh);
  return OK;
}


/* --- Init ---------------------------------------------------------- */

/** \brief Look for camera nodes and configure module.
 *
 * Reports:      OK
 *              S_viam_IEEE1394_CREATE_HANDLE
 */

ACTIVITY_EVENT
viamInitMain(int *report)
{
  viam_hardware_attach(vh);

  SDI_F->init_done = 1;
  return ETHER;
}


/* --- Configure ----------------------------------------------------- */

/** \brief Configure a bank for image acquisition
 *
 * codel EXEC of Configure
 *
 * \return EXEC END ETHER FAIL ZOMBIE
 */

ACTIVITY_EVENT
viamConfigureMain(ViamId *name, int *report)
{
  viam_bankdata_t *d;
  viam_bank_t bank;
  STATUS s;
  int i;

  bank = viam_bank_getbyname(vh, name->id);
  if (!bank) VIAM_RETURN_WITHERRNO(report, ETHER);
  d = viam_bank_usrdata(bank);
  if (d->pdata) { *report = S_viam_CONFIGURED; return ETHER; }

  /* connect to pom */
  s = viamPomInit(bank, report);
  if (s != OK) return ETHER;

  /* configure hardware */
  if (viam_bank_configure(vh, bank))
    VIAM_RETURN_WITHERRNO(report, ETHER);

  /* initialize poster */
  d->pdata = viamPosterCreate(vh, bank, d->buffers, report);
  if (!d->pdata) {
    viam_bank_unconfigure(vh, bank);
    return ETHER;
  }

  /* remember configured banks for delete hook */
  for(i=0; i<nconfigured; i++)
    if (configured[i] == bank) break;
  if (i >= nconfigured) {
    nconfigured++;
    configured = realloc(configured, nconfigured * sizeof(configured[0]));
    if (configured)
      configured[nconfigured-1] = bank;
  }

  return ETHER;
}


/* --- Reset --------------------------------------------------------- */

/** \brief Stop module and reset configuration info
 *
 * codel EXEC of Reset
 *
 * \returns  EXEC END ETHER FAIL ZOMBIE
 */

ACTIVITY_EVENT
viamResetMain(int *report)
{
  viamMainEnd();
  viamMainInit(report);
  return ETHER;
}


/* --- FilterList ---------------------------------------------------------- */

/** \brief EXEC codel of FilterList
 *
 * Reports:      OK
 *              S_viam_NO_SUCH_CAMERA
 *              S_viam_FWBUS
 *              S_viam_ARRAY_SIZE
 *
 * \return ETHER
 */

ACTIVITY_EVENT
viamFilterListMain(ViamId *name, ViamFilterList *list, int *report)
{
  viam_camera_t camera;
  unsigned int methods, automodes;
  unsigned int i, count;

  camera = viam_camera_getbyname(vh, name->id);
  if (!camera) VIAM_RETURN_WITHERRNO(report, ETHER);

  count = 0;
  for(i = 1; i < sizeof(list->type)/sizeof(list->type[0]); i++)
    switch(i) {
      case VIAM_FILTER_NONE:
      case VIAM_FILTER_FORMAT:
      case VIAM_FILTER_SUBSAMPLE:
      case VIAM_FILTER_SCALE:
#ifdef VIAM_OPENCV_SUPPORT
      case VIAM_FILTER_DISTORT:
      case VIAM_FILTER_DISTORT_NN:
      case VIAM_FILTER_RECTIFY:
      case VIAM_FILTER_RECTIFY_NN:
#endif /* VIAM_OPENCV_SUPPORT */
      case VIAM_FILTER_BRIGHTNESS:
      case VIAM_FILTER_EXPOSURE:
      case VIAM_FILTER_SHARPNESS:
      case VIAM_FILTER_GAIN:
      case VIAM_FILTER_SHUTTER:
      case VIAM_FILTER_WHITE_BALANCE:
      case VIAM_FILTER_HUE:
      case VIAM_FILTER_SATURATION:
      case VIAM_FILTER_GAMMA:
      case VIAM_FILTER_TEMPERATURE:
      case VIAM_FILTER_ZOOM:
      case VIAM_FILTER_FOCUS:
      case VIAM_FILTER_PAN:
      case VIAM_FILTER_TILT:
#ifdef VIAM_OPENCV_SUPPORT
      case VIAM_FILTER_LOG:
#endif /* VIAM_OPENCV_SUPPORT */
	if (viam_queryfeature(vh, camera, i, &methods, NULL)) continue;
	if (methods) list->type[count++] = i;
  }

  for(; count < sizeof(list->type)/sizeof(list->type[0]); count++)
    list->type[count++] = VIAM_FILTER_NONE;

  return ETHER;
}


/* --- FilterGetCap -------------------------------------------------- */

/** \brief EXEC codel of FilterGetCap
 *
 * Reports:      OK
 *              S_viam_NO_SUCH_CAMERA
 *              S_viam_FWBUS
 *              S_viam_ARRAY_SIZE
 *
 * \return ETHER
 */

ACTIVITY_EVENT
viamFilterGetCapMain(ViamFilterGetCap *getcap, ViamFilterCap *cap, int *report)
{
  viam_camera_t camera;
  unsigned int methods, automodes;
  unsigned int i, count;

  camera = viam_camera_getbyname(vh, getcap->camera.id);
  if (!camera) VIAM_RETURN_WITHERRNO(report, ETHER);

  if (viam_queryfeature(vh, camera, getcap->type, &methods, &automodes))
    VIAM_RETURN_WITHERRNO(report, ETHER);

  memset(cap, 0, sizeof(*cap));
  count = 0;
  for(i=1; i<0x10; i<<=1) if (methods & i) {
    if (count >= sizeof(cap->methods)/sizeof(cap->methods[0])) {
      *report = S_viam_ARRAY_SIZE;
      return ETHER;
    }

    cap->methods[count++] = i;
  }
  count = 0;
  for(i=1; i<0x10; i<<=1) if (automodes & i) {
    if (count >= sizeof(cap->automodes)/sizeof(cap->automodes[0])) {
      *report = S_viam_ARRAY_SIZE;
      return ETHER;
    }

    cap->automodes[count++] = i;
  }

  return ETHER;
}


/* --- viamCameraListModesMain --------------------------------------- */

/** \brief EXEC codel of CameraListModes
 *
 * List available acquisition modes
 * Reports: OK, S_viam_FWBUS
 *
 * \return ETHER
 */

ACTIVITY_EVENT
viamCameraListHWModesMain(ViamHWModeListOffset *hwmodeoffset,
			  ViamHWModeList *hwmodes, int *report)
{
  viam_camera_t camera;
  int i;

  camera = viam_camera_getbyname(vh, hwmodeoffset->name.id);
  if (!camera) VIAM_RETURN_WITHERRNO(report, ETHER);

  hwmodes->count =
    viam_queryhwmodes(vh, camera, hwmodeoffset->offset, hwmodes->modes,
		      sizeof(hwmodes->modes)/sizeof(hwmodes->modes[0]));

  /* reset extra elements */
  for(i = hwmodes->count;
      i<sizeof(hwmodes->modes)/sizeof(hwmodes->modes[0]); i++) {
    hwmodes->modes[i].size = VIAM_HWSZ_INVALID;
    hwmodes->modes[i].format = VIAM_HWFMT_INVALID;
    hwmodes->modes[i].crop = VIAM_HW_FIXED;
    hwmodes->modes[i].fps = VIAM_HWFPS_INVALID;
    hwmodes->modes[i].trigger = VIAM_HWTRIGGER_INVALID;
  }

  return ETHER;
}


/* --- viamDisplayMain ----------------------------------------------------- */

/** \brief EXEC codel of Display
 *
 * Report: OK, S_viam_OPENCV
 *
 * \return OK or ERROR
 */

ACTIVITY_EVENT
viamDisplayMain(ViamDisplay *display, int *report)
{
#ifdef VIAM_DISPLAY
  viam_bank_t bank;
  viam_image_ref image;
  viam_imagedata_t *idata;
  int i, s;

  /* find bank */
  bank = viam_bank_getbyname(vh, display->bank.id);
  if (!bank) VIAM_RETURN_WITHERRNO(report, ERROR);
  if (!viam_bank_isconfigured(bank)) {
    *report = S_viam_NOT_CONFIGURED;
    return ETHER;
  }

  /* find image */
  for(i=0; i<viam_bank_nimages(bank); i++) {
    image = viam_bank_image(vh, bank, i);
    if (!strcmp(viam_image_name(image), display->image.id))
      break;
    image = NULL;
  }
  if (!image) { *report = S_viam_NO_SUCH_IMAGE; return ETHER; }
  idata = viamIData(bank, i);
  if (!idata) { *report = S_viam_MEMORY; return ETHER; }

  /* create/destroy window */
  if (idata->window) {
    char *w = idata->window;

    idata->window = NULL;
    cvDestroyWindow(w);
    free(w);
    cvWaitKey(1); /* nice feature of opencv not destroying windows */
    cvWaitKey(1);
    cvWaitKey(1);
  }

  if (display->enable == VIAM_ON) {
    viam_geometry_t geom;
    char *w;

    s = viam_image_geometry(vh, image, &geom);
    if (s) VIAM_RETURN_WITHERRNO(report, ETHER);

    w = malloc(
      strlen(display->bank.id) + strlen(" - ") + strlen(display->image.id)
      + 1);
    if (!w) { *report = S_viam_NO_SUCH_IMAGE; return ETHER; }
    sprintf(w, "%s - %s", display->bank.id, display->image.id);

    cvNamedWindow(w, 0);
    if (cvGetErrStatus()) { *report = S_viam_OPENCV; return ETHER; }

    if (display->width == 0 && display->height == 0) {
      display->width = geom.width;
      display->height = geom.height;
    }
    if (display->width == 0)
      display->width = geom.width * display->height / geom.height;
    if (display->height == 0)
      display->height = geom.height * display->width / geom.width;
    cvResizeWindow(w, display->width, display->height);
    idata->wwidth = display->width;
    idata->wheight = display->height;

    cvWaitKey(1); /* nice feature of opencv not creating windows */
    cvWaitKey(1);
    cvWaitKey(1);
    idata->window = w;
  }

  /* store misc parameters */
  idata->vtags = display->vtag;

  return ETHER;
#else
  *report = S_viam_FEATURE_DISABLED;
  return ETHER;
#endif /* VIAM_DISPLAY */
}
